local addonName = "ProfessionsVault"
ProfessionsVault = LibStub("AceAddon-3.0"):NewAddon(addonName)
local addon = ProfessionsVault 
local AceGUI = LibStub("AceGUI-3.0")
local revision = tonumber(("$Revision: 1 $"):match("%d+"))

local defaults = {
  profile = {
    debug = false, -- for addon debugging
    autoclose = false, -- close win after using a link
    pos = {
      ["height"] = 432,
      ["width"]  = 460,
      ["left"]   = 350,
      ["top"]    = 823,
    }
  }
}

local headerfont = GameFontHighlightLarge
local bodyfont = GameFontHighlightLarge
local profpages = false -- show a page for each profession
local activateontree = true


local primaryProf = { 
	["Alchemy"] = true, 
	["Blacksmithing"] = true, 
	["Enchanting"] = true, 
	["Engineering"] = true, 
	-- ["Herbalism"] = true, 
	["Inscription"] = true, 
	["Jewelcrafting"] = true, 
	["Leatherworking"] = true, 
	["Smelting"] = true, 
	["Tailoring"] = true, 
}
local secondaryProf = { 
	["Archaeology"] = true, 
	["Cooking"] = true, 
	["First Aid"] = true, 
	-- ["Fishing"] = true 
}
local allProf = {}
for p,v in pairs(primaryProf) do allProf[p]=v end
for p,v in pairs(secondaryProf) do allProf[p]=v end

local DB, DBc, settings
local charName

local function chatMsg(msg)
     DEFAULT_CHAT_FRAME:AddMessage(addonName..": "..msg)
end
local function debug(msg)
  if settings.debug then
     chatMsg(msg)
  end
end

local function myOptions() 
return {
  type = "group",
  args = {
    debug = {
      name = "Debug",
      desc = "Toggle debugging output",
      type = "toggle",
      set = function(info,val) settings.debug = val end,
      get = function(info) return settings.debug end
    },
    reset = {
      name = "Reset",
      desc = "Reset the database",
      type = "execute",
      func = function() chatMsg("Reset complete."); addon.db:ResetDB()  end,
    },
    dump = {
      name = "Dump",
      desc = "Dump the database to chat window",
      type = "execute",
      func = function() addon:Dump()  end,
    },
    show = {
      name = "Show",
      desc = "Show/Hide the ProfessionsVault window",
      type = "execute",
      func = function() addon:ToggleWindow()  end,
    },
    autoclose = {
      name = "Autoclose",
      desc = "Automatically close the ProfessionsVault window when using a link",
      type = "toggle",
      set = function(info,val) settings.autoclose = val end,
      get = function(info) return settings.autoclose end
    },
  }
} 
end

function addon:RefreshConfig()
  debug("RefreshConfig")
  charName = UnitName("player")
  DB = self.db.realm
  DB.chars = DB.chars or {}
  DB.chars[charName] = DB.chars[charName] or {}
  DBc = DB.chars[charName]
  settings = addon.db.profile
end

function addon:OnInitialize()
  self.db = LibStub("AceDB-3.0"):New("ProfessionsVaultDB", defaults)
  --LibStub("AceConfigRegistry-3.0"):ValidateOptionsTable(myOptions(), addonName)
  LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, myOptions(), {"professionsvault", "pv"})
  LibStub("AceEvent-3.0"):Embed(addon)
  settings = addon.db.profile

  debug("OnInitialize")

  self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
  self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
  self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
  self.db.RegisterCallback(self, "OnDatabaseReset", "RefreshConfig")
  self:RefreshConfig()
end

local LDB
function addon:OnEnable()
  debug("OnEnable")
  self:RegisterEvent("TRADE_SKILL_SHOW")

  --if not next(DBc) then -- no professions registered on this toon
  --  addon:ScanProfessions()
  --end

  if LDB then
    return
  end
  LDB = LibStub:GetLibrary("LibDataBroker-1.1",true)
  if LDB then
    LDB:NewDataObject(addonName, {
        type = "launcher",
        label = addonName,
        icon = "Interface\\Icons\\spell_shadow_mindtwisting",
        OnClick = function(self, button)
                if button == "RightButton" then
                        addon:OpenMenu(self,addon)
                else
                        addon:ToggleWindow()
                end
        end,
        OnTooltipShow = function(tooltip)
                if tooltip and tooltip.AddLine then
                        tooltip:SetText(addonName)
                        tooltip:AddLine("|cffff8040Left Click|r to toggle the ProfessionsVault window")
                        tooltip:AddLine("|cffff8040Right Click|r for menu")
                        tooltip:Show()
                end
        end,
     })
  end 
end

function addon:TRADE_SKILL_SHOW()
  debug("TRADE_SKILL_SHOW")
  addon:UpdateTrade()
end

function addon:UpdateTrade()
  local name, rank, rankmax = GetTradeSkillLine()
  local link = GetTradeSkillListLink()
  local linked, linkedName = IsTradeSkillLinked()
  if (linked) then
    debug("Ignoring "..linkedName.."'s "..name);
    return
  end
  if (name and name ~= "UNKNOWN" and link) then
    DBc[name] = {}

    DBc[name].link = link
    DBc[name].rank = rank
    DBc[name].rankmax = rankmax
    DBc[name].icon = GetTradeSkillTexture()
    DBc[name].faction = UnitFactionGroup("player")
    debug("Updated "..charName.."'s "..name);
  end
end

-- traversal function for character table
local cnext_sorted_names = {}
local function cnext(t,i)
   -- return them in reverse order
   if #cnext_sorted_names == 0 then 
     return nil
   else
      local n = cnext_sorted_names[#cnext_sorted_names]
      table.remove(cnext_sorted_names, #cnext_sorted_names)
      return n, t[n]
   end
end
local function cpairs(t)
  cnext_sorted_names = {}
  for n,_ in pairs(t) do
    table.insert(cnext_sorted_names, n)
  end
  table.sort(cnext_sorted_names, function (a,b) return b == charName or (a ~= charName and a > b) end)
  --myprint(cnext_sorted_names)
  return cnext, t, nil
end

function addon:Dump()
  for cname,dbc in cpairs(DB.chars) do
    for pname,pinfo in pairs(dbc) do
       local rank = pinfo.rank or "??"
       local rankmax = pinfo.rankmax or "??"
       chatMsg(cname.."'s "..pinfo.link.."   ("..rank.."/"..rankmax..")")
    end
  end
end

local function OnClose(widget)
  debug("OnClose")
  AceGUI:Release(widget)
  addon.gui = nil
end

local function cleanlink(link)
  local res = string.match(link,"|H(trade:[^|]+)|h")
  debug("cleanlink("..link..") => ".. res)
  return res
end

local function ActivateLink(cname,pname,link)
  debug("ActivateLink: "..cname.." "..pname)
  if IsShiftKeyDown() then
    local activeEditBox =  ChatEdit_GetActiveWindow();
    if activeEditBox then
      if (not ChatEdit_InsertLink(link) ) then
         debug("Failed to add tradeskill link to "..activeEditBox:GetName())
	 return
      end
    else
      return
    end
  else
    if cname == charName then
      CastSpellByName(pname)
    else
      local name, rank, rankmax = GetTradeSkillLine()
      local linked, linkedName = IsTradeSkillLinked()
      if linked and linkedName == cname and name and name == pname then -- already open so close it
        CloseTradeSkill()
      else -- open it (undocumented API - sshhh)
        SetItemRef(cleanlink(link),link,"LeftButton",ChatFrame1)
      end
    end
  end
  if settings.autoclose then
    addon.gui:Hide()
  end
end

function addon:ScanProfessions()
  debug("ScanProfessions")
  CloseTradeSkill()
  for p in pairs(allProf) do
    CastSpellByName(p) 
    CloseTradeSkill()
  end
  chatMsg("Profession Scan complete!")
  if addon.gui then
    addon:ToggleWindow()
    addon:ToggleWindow()
  end
end

local function ProfIcon(cname,pname,pinfo)
  local icon = AceGUI:Create("Icon")
  icon:SetImage(pinfo.icon)
  icon:SetCallback("OnClick", function(button) ActivateLink(cname,pname,pinfo.link) end)
  icon:SetCallback("OnEnter", function(button) GameTooltip:SetOwner(button.frame, "ANCHOR_CURSOR"); 
                                               GameTooltip:SetText(pname.." ("..pinfo.rank.." / "..pinfo.rankmax..")"); 
					       GameTooltip:Show() end)
  icon:SetCallback("OnLeave", function(button) GameTooltip:Hide() end)
  return icon
end

local function MakeButton(text, fn) 
  local button = AceGUI:Create("Button")
  button:SetText(text)
  button:SetWidth(200)
  button:SetCallback("OnClick",fn)
  return button
end

local function OnSelect(widget, event, value)
  local proftree, _, cname, pname = string.find(value,"^([%w%s]+)\001([%w%s]+)$")
  if proftree then
    debug("OnSelect: "..cname.." / "..pname)
    addon.gui:SetStatusText(cname.." / "..pname)
    settings.treepath = { cname , pname }
  else
    cname = value
    debug("OnSelect: "..cname)
    addon.gui:SetStatusText(cname)
    settings.treepath = { cname }
  end
  if not proftree or not settings.profpages then
    widget:ReleaseChildren()
    widget:SetLayout("List") 
    -- body for a character
    local label = AceGUI:Create("Label")
    label:SetText(cname.."'s Professions\n")
    label:SetFullWidth(true)
    label:SetFontObject(headerfont)
    widget:AddChild(label)
    if cname == charName and not next(DBc) then
      widget:AddChild(MakeButton("Scan My Professions", function () addon:ScanProfessions(); end))
      return
    end
    local scroller = AceGUI:Create("ScrollFrame")
    scroller:SetLayout("List")
    scroller:SetFullHeight(true)
    scroller:SetFullWidth(true)
    widget:AddChild(scroller)
    for i=1,2 do
      local label = AceGUI:Create("Label")
      label:SetText(i==1 and "Primary:\n" or "Secondary:\n")
      label:SetFontObject(bodyfont)
      scroller:AddChild(label)
      local body = AceGUI:Create("SimpleGroup")
      body:SetLayout("Flow")
      body:SetFullWidth(true)
      --body:SetAutoAdjustHeight(true)
      for pname, pinfo in pairs(DB.chars[cname]) do
        if (i==1 and not secondaryProf[pname]) or (i==2 and secondaryProf[pname]) then
          body:AddChild(ProfIcon(cname,pname,pinfo))
	end
      end
      scroller:AddChild(body)
    end
    if proftree and activateontree then
      ActivateLink(cname,pname,DB.chars[cname][pname].link)
    end
    return
  end

 
  local pinfo = DB.chars[cname][pname]

  widget:ReleaseChildren()
  widget:SetLayout("List") 
  
  -- body for a particular profession
  local header = AceGUI:Create("SimpleGroup")
  header:SetFullWidth(true)
  header:SetLayout("Flow") 
  header:AddChild(ProfIcon(cname,pname,pinfo))
  local label = AceGUI:Create("Label")
  label:SetText(cname.."'s "..pname.."\n")
  label:SetFontObject(headerfont)
  header:AddChild(label)
  widget:AddChild(header)

  local body = AceGUI:Create("SimpleGroup")
  body:SetFullWidth(true)
  local skill = AceGUI:Create("Label")
  skill:SetText("Skill level: "..pinfo.rank.." / "..pinfo.rankmax.."\n")
  skill:SetFontObject(bodyfont)
  skill:SetFullWidth(true)
  body:AddChild(skill)
  local link = AceGUI:Create("InteractiveLabel")
  link:SetCallback("OnClick", function(button) ActivateLink(cname,pname,pinfo.link) end)
  link:SetText("Skill link: "..pinfo.link)
  link:SetFontObject(bodyfont)
  link:SetFullWidth(true)
  body:AddChild(link)
  widget:AddChild(body)
end

function addon:CreateWindow()
  if addon.gui then
    return
  end
  local f = AceGUI:Create("Frame")
  addon.gui = f
  f:SetTitle("Professions Vault")
  f:SetStatusText("ProfessionsVault listing of your toon professions")
  f:SetCallback("OnClose", OnClose)
  f:SetLayout("Flow")
  f.frame:SetClampedToScreen(true)
  settings.pos = settings.pos or {}
  f:SetStatusTable(settings.pos)
  f:SetAutoAdjustHeight(true)

  f:AddChild(MakeButton("Profession Pane", function () ToggleSpellBook(BOOKTYPE_PROFESSION); end)) 
  f:AddChild(MakeButton("Scan My Professions", function () addon:ScanProfessions(); end))

  local treedata = {}
  for cname,dbc in cpairs(DB.chars) do
    local ct = { value = cname, text = cname, 
                 icon = "Interface\\GossipFrame\\ActiveQuestIcon",
                 children = {} }
    --print(cname)
    for i=1,2 do
      for pname,pinfo in pairs(dbc) do
        if pinfo.faction == "Horde" then ct.icon = "Interface\\TargetingFrame\\UI-PVP-Horde" 
        elseif pinfo.faction == "Alliance" then ct.icon = "Interface\\TargetingFrame\\UI-PVP-Alliance" end
        if (i==1 and not secondaryProf[pname]) or (i==2 and secondaryProf[pname]) then
          table.insert(ct.children, { value = pname, icon = pinfo.icon,
                                      text = pname.." ("..pinfo.rank.." / "..pinfo.rankmax..")" })
	end
      end
    end
    table.insert(treedata, ct)
  end

  local tree = AceGUI:Create("TreeGroup")
  tree:SetFullWidth(true)
  tree:SetFullHeight(true)
  --tree:SetAutoAdjustHeight(true)
  tree:SetTree(treedata)
  tree:SetCallback("OnGroupSelected", OnSelect)
  tree:EnableButtonTooltips(false)
  f:AddChild(tree)
  if settings.treepath then
    local oldact = activateontree
    activateontree = false
    tree:SelectByPath(unpack(settings.treepath))
    activateontree = oldact
  end
  settings.treestatus = settings.treestatus or {}
  tree:SetStatusTable(settings.treestatus)
end

function addon:ToggleWindow()
  debug("ToggleWindow")

  if addon.gui then
    addon.gui:Hide()
  else
    addon:CreateWindow()
  end
end

